# def list_plugins():
#    path = os.path.dirname(__file__)
#    for x in os.listdir(path):
#        if x != "__init__.py" and x.endswith(".py"):
#            print(x)
#            module = __import__(x)
#            print(help(module))
#
#
# list_plugins()
